% *** WLS-VB Algorithm ***
% *high-dimensional VB algorithm for regular prior and non-homogenous noise
%  to solve set of equations m = Ax + e, where x is sparse vector of unknowns
% *see paper 'Comparison of sparse Bayesian regularization methods 
%  for high-dimensional inverse problem'
% *'visualisation of the reconstruction' is prepared for case, 
%  where x is n*n image - a modification is easy and possible
%
% % written by Jakub Sevcik and Vaclav Smidl 2016

% Load the measurement matrix (A) and the vector (m)
% *a training dataset is downloadable from http://www.fips.fi/dataset.php
load Data328 A m  
m = m(:);

% % WLS-VB Algorithm
% inicialization
iterations = 100;

[M,N] = size(A);
n = sqrt(N);

alfa = 10;
gammaa0 = 1e-8;
deltaa0 = 1e-8;
gammaB0 = 1e-8*ones(M,1);
deltaB0 = 1e-8*ones(M,1);
beta = ones(M,1);

% calculate Ap2t = (A.^2)' for faster computation
Ap2t = sparse(A.^2)';

% iterative section
for iter = 1:1:iterations
    diSb = Ap2t*beta + alfa*ones(N,1); % idSb = diag(Atdiag(beta)A+alfa*I)
    dSb = 1./diSb;

    fun = @(x) A'*(beta.*(A*x))+ alfa.*x;
    b = A'*(beta.*m);
    [x,flag] = pcg(fun,b,[],50);
    %flags(iter) = flag;
    %X(iter)= x;
    %x(x<0) = 0.0*x(x<0);
    
    gammaa = gammaa0 + 0.5*N;
    deltaa = deltaa0 + 0.5*x'*x + 0.5*sum(dSb);
    alfa = gammaa/deltaa;
    
    gammaB = gammaB0 + 0.5;
    deltaB = deltaB0 + 0.5*m.^2 - m.*(A*x) + 0.5*(A*x).^2 + 0.5*Ap2t'*dSb;
    beta = gammaB./deltaB;
    
    if mod(iter,10) == 0
            disp([iter, iterations])
    end
end

% % visualisation of the reconstruction
figure
imagesc(reshape(x,n,n))
colormap gray
axis square
axis off
title(['WLS-VB Algorithm with ',num2str(iterations),' iterations'])
